@@ -35,6 +35,29 @@ using namespace vfalib;
 
 namespace {
 
+enum { /* see libXfont */
+	PCF_PROPERTIES       = 0x1U,
+	PCF_ACCELERATORS     = 0x2U,
+	PCF_METRICS          = 0x4U,
+	PCF_BITMAPS          = 0x8U,
+	PCF_INK_METRICS      = 0x10U,
+	PCF_BDF_ENCODINGS    = 0x20U,
+	PCF_SWIDTHS          = 0x40U,
+	PCF_GLYPH_NAMES      = 0x80U,
+	PCF_BDF_ACCELERATORS = 0x100U,
+};
+
+enum {
+	PCF_DEFAULT_FORMAT     = 0,
+	PCF_INKBOUNDS          = 0x200U,
+	PCF_ACCEL_W_INKBOUNDS  = 0x100U,
+	PCF_COMPRESSED_METRICS = 0x100U,
+
+	PCF_GLYPH_PAD_MASK     = 0x3U,
+	PCF_BYTE_MASK          = 0x4U,
+	PCF_FORMAT_MASK        = 0xffffff00U,
+};
+
 enum {
 	PSF1_MAGIC0 = 0x36,
 	PSF1_MAGIC1 = 0x04,
@@ -62,6 +85,10 @@ struct deleter {
 	void operator()(HXdir *d) { HXdir_close(d); }
 };
 
+struct pcf_table {
+	uint32_t type, format, size, offset;
+};
+
 struct psf2_header {
 	uint8_t magic[4];
 	uint32_t version, headersize, flags, length, charsize, height, width;
@@ -292,6 +319,12 @@ void font::lgeuf()
 		m_glyph[it->second].lge();
 }
 
+void font::overstrike(unsigned int px)
+{
+	for (auto &g : m_glyph)
+		g = g.overstrike(px);
+}
+
 struct bdfglystate {
 	int uc = -1, w = 0, h = 0, of_left = 0, of_baseline = 0;
 	unsigned int dwidth = 0, lr = 0;
@@ -353,7 +386,38 @@ static glyph bdfcomplete(const bdfglystate &cchar)
 	dst_rect.y = std::max(0, static_cast<int>(cchar.font_ascent) - cchar.of_baseline - cchar.h);
 	dst_rect.w = cchar.dwidth;
 	dst_rect.h = cchar.font_height;
-	return g.copy_to_blank(src_rect, dst_rect);
+	return g.copy_rect_to(src_rect, glyph(dst_rect), dst_rect);
+}
+
+#include "glynames.cpp"
+
+static std::string translate_charname(const char *s)
+{
+	if (*s == 'C') {
+		const char *p;
+		for (p = s; HX_isdigit(*p) && *p != '\0'; ++p)
+			;
+		if (*p == '\0')
+			return s;
+	}
+	if (strncmp(s, "uni", 3) == 0) {
+		char *end = nullptr;
+		auto uc = strtoul(&s[1], &end, 16);
+		if (end != nullptr && *end == '\0') {
+			char buf[16];
+			snprintf(buf, sizeof(buf), "C%lu", uc);
+			return buf;
+		}
+	}
+	auto it = std::lower_bound(std::cbegin(ff_glyph_names), std::cend(ff_glyph_names), s,
+	          [](const std::pair<const char *, uint32_t> &p, const char *q) {
+	          	return strcmp(p.first, q) < 0;
+	          });
+	if (it == std::cend(ff_glyph_names) || strcmp(it->first, s) != 0)
+		return s;
+	char buf[16];
+	snprintf(buf, sizeof(buf), "C%u", it->second);
+	return buf;
 }
 
 int font::load_bdf(const char *filename)
@@ -371,8 +435,9 @@ int font::load_bdf(const char *filename)
 	bdfglystate cchar;
 
 	while (HX_getl(&line, fp.get()) != nullptr) {
+		HX_chomp(line);
 		if (state == BDF_NONE) {
-			if (strncmp(line, "STARTFONT 2.1\n", 14) == 0) {
+			if (strcmp(line, "STARTFONT 2.1") == 0) {
 				state = BDF_FONT;
 				continue;
 			}
@@ -382,7 +447,7 @@ int font::load_bdf(const char *filename)
 			if (strncmp(line, "STARTCHAR ", 10) == 0) {
 				cchar.reset();
 				cchar.font_height = cchar.font_ascent + cchar.font_descent;
-				cchar.name = line + 10;
+				cchar.name = translate_charname(line + 10);
 				state = BDF_CHAR;
 				continue;
 			}
@@ -418,7 +483,7 @@ int font::load_bdf(const char *filename)
 				cchar.lr = cchar.h;
 				continue;
 			}
-			if (strcmp(line, "BITMAP\n") == 0) {
+			if (strcmp(line, "BITMAP") == 0) {
 				state = cchar.lr == 0 ? BDF_PASTBITMAP : BDF_BITMAP;
 				continue;
 			}
@@ -434,7 +499,7 @@ int font::load_bdf(const char *filename)
 				continue;
 			}
 		} else if (state == BDF_PASTBITMAP) {
-			if (strcmp(line, "ENDCHAR\n") == 0) {
+			if (strcmp(line, "ENDCHAR") == 0) {
 				if (cchar.uc != -1) {
 					m_unicode_map->add_i2u(m_glyph.size(), cchar.uc);
 					m_glyph.push_back(bdfcomplete(std::move(cchar)));
@@ -582,6 +647,133 @@ int font::load_hex(const char *file)
 	return 0;
 }
 
+static int load_pcf_props(FILE *fp, std::map<std::string, std::string> &map)
+{
+	uint32_t val = 0;
+	if (fread(&val, 4, 1, fp) != 1)
+		return -EINVAL;
+	val = le32_to_cpu(val);
+	if ((val & PCF_FORMAT_MASK) != (PCF_DEFAULT_FORMAT & PCF_FORMAT_MASK))
+		return -EINVAL;
+	bool be = val & PCF_BYTE_MASK;
+	if (fread(&val, 4, 1, fp) != 1)
+		return -EINVAL;
+	auto numprop = be ? be32_to_cpu(val) : le32_to_cpu(val);
+	auto tbl_offset = ftell(fp);
+	fseek(fp, numprop * 9, SEEK_CUR);
+	fseek(fp, 4 - (ftell(fp) & 3), SEEK_CUR);
+	if (fread(&val, 4, 1, fp) != 1)
+		return -EINVAL;
+	val = be ? be32_to_cpu(val) : le32_to_cpu(val);
+	std::string sblk;
+	sblk.resize(val);
+	if (fread(&sblk[0], sblk.size(), 1, fp) != 1)
+		return -EINVAL;
+	fseek(fp, tbl_offset, SEEK_SET);
+	for (unsigned int i = 0; i < numprop; ++i) {
+		if (fread(&val, 4, 1, fp) != 1)
+			return -EINVAL;
+		auto name_idx = be ? be32_to_cpu(val) : le32_to_cpu(val);
+		uint8_t is_string = 0;
+		if (fread(&is_string, 1, 1, fp) != 1)
+			return -EINVAL;
+		if (fread(&val, 4, 1, fp) != 1)
+			return -EINVAL;
+		auto val_idx = be ? be32_to_cpu(val) : le32_to_cpu(val);
+		if (name_idx >= sblk.size() || val_idx >= sblk.size())
+			return -EINVAL;
+		if (is_string)
+			map[&sblk[name_idx]] = &sblk[val_idx];
+		else
+			map[&sblk[name_idx]] = std::to_string(val_idx);
+	}
+	return 0;
+}
+
+static int load_pcf_bitmaps(FILE *fp)
+{
+	uint32_t val;
+	if (fread(&val, 4, 1, fp) != 1)
+		return -EINVAL;
+	auto fmt = le32_to_cpu(val);
+	if ((fmt & PCF_FORMAT_MASK) != (PCF_DEFAULT_FORMAT & PCF_FORMAT_MASK))
+		return -EINVAL;
+	bool be = fmt & PCF_BYTE_MASK;
+	if (fread(&val, 4, 1, fp) != 1)
+		return -EINVAL;
+	auto numbitmaps = be ? be32_to_cpu(val) : le32_to_cpu(val);
+	static constexpr unsigned int glypadopts = 4;
+	std::vector<uint32_t> offsets(numbitmaps);
+	uint32_t bmpsize[glypadopts];
+	if (fread(&offsets[0], 4, numbitmaps, fp) != numbitmaps ||
+	    fread(&bmpsize[0], 4, glypadopts, fp) != glypadopts)
+		return -EINVAL;
+	for (unsigned int i = 0; i < numbitmaps; ++i) {
+		offsets[i] = be ? be32_to_cpu(offsets[i]) : le32_to_cpu(offsets[i]);
+		printf("bmp %u offset %u\n",
+			i, offsets[i]);
+	}
+	for (unsigned int i = 0; i < glypadopts; ++i) {
+		bmpsize[i] = be ? be32_to_cpu(bmpsize[i]) : le32_to_cpu(bmpsize[i]);
+		printf("padopt %u size %u\n", i, bmpsize[i]);
+	}
+	std::string bmapbuf;
+	#define PCF_GLYPH_PAD_INDEX(f) ((f) & PCF_GLYPH_PAD_MASK)
+	bmapbuf.resize(bmpsize[PCF_GLYPH_PAD_INDEX(fmt)]);
+	fprintf(stderr, "buf %zu\n", bmapbuf.size());
+	if (fread(&bmapbuf[0], bmapbuf.size(), 1, fp) != 1)
+		return -EINVAL;
+	return 0;
+}
+
+int font::load_pcf(const char *filename)
+{
+	std::unique_ptr<FILE, deleter> fp(vfopen(filename, "r"));
+	if (fp == nullptr)
+		return -errno;
+
+	uint32_t val = 0;
+	if (fread(&val, 4, 1, fp.get()) != 1)
+		return -EINVAL;
+	if (memcmp(&val, "\x01""fcp", 4) != 0)
+		return -EINVAL;
+	if (fread(&val, 4, 1, fp.get()) != 1)
+		return -EINVAL;
+	val = le32_to_cpu(val);
+	std::vector<pcf_table> table(val);
+	const pcf_table *bmp_table = nullptr, *prop_table = nullptr;
+	for (size_t i = 0; i < val; ++i) {
+		auto &t = table[i];
+		if (fread(&t, sizeof(t), 1, fp.get()) != 1)
+			return -EINVAL;
+		t.type   = le32_to_cpu(t.type);
+		t.format = le32_to_cpu(t.format);
+		t.size   = le32_to_cpu(t.size);
+		t.offset = le32_to_cpu(t.offset);
+		if (t.type == PCF_PROPERTIES)
+			prop_table = &t;
+		else if (t.type == PCF_BITMAPS)
+			bmp_table = &t;
+		fprintf(stderr, "Table %zu: type %xh format %xh size %u offset %u\n",
+			i, t.type, t.format, t.size, t.offset);
+	}
+	if (prop_table == nullptr || fseek(fp.get(), prop_table->offset, SEEK_SET) != 0) {
+		fprintf(stderr, "pcf: no properties\n");
+		return -EINVAL;
+	}
+	std::map<std::string, std::string> propmap;
+	auto ret = load_pcf_props(fp.get(), propmap);
+	if (ret != 0)
+		return ret;
+	//y=PIXEL_SIZE,x=QUAD_WIDTH
+	if (bmp_table == nullptr || fseek(fp.get(), bmp_table->offset, SEEK_SET) != 0) {
+		fprintf(stderr, "pcf: no bitmaps\n");
+		return -EINVAL;
+	}
+	load_pcf_bitmaps(fp.get());
+	return 0;
+}
+
 static char32_t nextutf8(FILE *fp)
 {
 	unsigned int nbyte = 0;
@@ -966,8 +1158,10 @@ static void name_reminder(font::propmap_t &props)
 	auto z = c.empty() || c == "vfontas output";
 	if (x && y && z) {
 		fprintf(stderr, "Hint: Consider -setname <name>, "
-		        "or the detailed version, "
-		        "-setprop {FontName|FamilyName|FullName}.\n");
+		        "or the detailed version, e.g.\n"
+		        "\t-setprop FontName aerial_20  # PostScript name\n"
+		        "\t-setprop FamilyName \"Aerial 2.0\"\n"
+		        "\t-setprop FullName \"Aerial 2.0 Bold\"\n");
 		return;
 	}
 	if (x)
@@ -1615,9 +1809,10 @@ glyph glyph::create_from_rpad(const vfsize &size, const char *buf, size_t z)
 	return ng;
 }
 
-glyph glyph::copy_rect(const vfrect &sof, const vfrect &pof) const
+glyph glyph::copy_rect_to(const vfrect &sof, const glyph &other,
+    const vfrect &pof, bool overwrite) const
 {
-	glyph ng = *this;
+	glyph out = other;
 
 	for (unsigned int y = sof.y; y < sof.y + sof.h && y < m_size.h; ++y) {
 		for (unsigned int x = sof.x; x < sof.x + sof.w && x < m_size.w; ++x) {
@@ -1627,34 +1822,14 @@ glyph glyph::copy_rect(const vfrect &sof, const vfrect &pof) const
 			    static_cast<unsigned int>(oy) >= pof.h)
 				continue;
 			bitpos ipos = y * m_size.w + x;
-			bitpos opos = oy * ng.m_size.w + ox;
+			bitpos opos = oy * out.m_size.w + ox;
 			if (m_data[ipos.byte] & ipos.mask)
-				ng.m_data[opos.byte] |= opos.mask;
-			else
-				ng.m_data[opos.byte] &= ~opos.mask;
+				out.m_data[opos.byte] |= opos.mask;
+			else if (overwrite)
+				out.m_data[opos.byte] &= ~opos.mask;
 		}
 	}
-	return ng;
-}
-
-glyph glyph::copy_to_blank(const vfrect &sof, const vfrect &pof) const
-{
-	glyph ng(pof);
-
-	for (unsigned int y = sof.y; y < sof.y + sof.h && y < m_size.h; ++y) {
-		for (unsigned int x = sof.x; x < sof.x + sof.w && x < m_size.w; ++x) {
-			int ox = pof.x + x - sof.x;
-			int oy = pof.y + y - sof.y;
-			if (ox < 0 || oy < 0 || static_cast<unsigned int>(ox) >= pof.w ||
-			    static_cast<unsigned int>(oy) >= pof.h)
-				continue;
-			bitpos ipos = y * m_size.w + x;
-			bitpos opos = oy * ng.m_size.w + ox;
-			if (m_data[ipos.byte] & ipos.mask)
-				ng.m_data[opos.byte] |= opos.mask;
-		}
-	}
-	return ng;
+	return out;
 }
 
 int glyph::find_baseline() const
@@ -1716,6 +1891,15 @@ void glyph::lge(unsigned int adj)
 	}
 }
 
+glyph glyph::overstrike(unsigned int px) const
+{
+	glyph composite(m_size);
+	for (unsigned int x = 0; x <= px; ++x)
+		composite = copy_rect_to(vfpos(0, 0) | m_size, composite,
+		            vfpos(x, 0) | m_size, false);
+	return composite;
+}
+
 std::string glyph::as_pbm() const
 {
 	auto bpg = bytes_per_glyph(m_size);
